#!/usr/bin/env python3

import argparse
import logging
import os

import logzero

from .bin.wdldoc import parse_file, traverse_directory


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Generate clean WDL documentation from source."
    )
    parser.add_argument(
        "sources",
        nargs="+",
        help="Top level directories to search for WDL files, "
        "or the WDL files themselves.",
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output_directory",
        help="Directory to store markdown files. Default is `./documentation`",
        default="./documentation",
        type=str,
    )
    parser.add_argument(
        "-d",
        "--description",
        help="If parameter meta fields use a JSON object, the key "
        "for the field containing the input description. "
        "Default is 'help'. Ignored if only strings are used.",
        default="help",
        type=str,
    )
    parser.add_argument(
        "-c",
        "--choices",
        help="If parameter meta fields use a JSON object, the key "
        "for the field containing the input choices. "
        "Default is 'choices'. Ignored if only strings are used.",
        default="choices",
        type=str,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Sets the log level to INFO.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--debug",
        help="Sets the log level to DEBUG.",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()

    logzero.loglevel(logging.WARNING)
    if args.verbose:
        logzero.loglevel(logging.INFO)
    if args.debug:
        logzero.loglevel(logging.DEBUG)

    keys = {"description": args.description, "choices": args.choices}

    for source in args.sources:
        if os.path.isdir(source):
            traverse_directory(source, args.output_directory, keys)
        elif os.path.isfile(source):
            parse_file(source, args.output_directory, keys)
        else:
            raise RuntimeError(f"Could not find {source}.")



import glob
import os
import pathlib
import sys
from typing import Dict, TextIO, Union

import logzero
import WDL as wdl

from .. import classify_inputs
from ..miniwdl.sources import read_source
from ..templates.markdown import MarkDownNode


def document_node(
    node: Union[wdl.Workflow, wdl.Task], doc: MarkDownNode, keys: Dict[str, str]
) -> None:
    doc.set_title(node.name)
    parameter_metadata = node.parameter_meta
    inputs = classify_inputs(node)
    doc.generate_inputs(inputs, parameter_metadata, keys)
    doc.generate_outputs(node.effective_outputs)
    doc.generate_meta(node.meta)


def document_workflow(
    document: wdl.Document, output: TextIO, keys: Dict[str, str]
) -> None:
    doc = MarkDownNode()
    doc.generate_header(document.source_text)
    document_node(document.workflow, doc, keys)
    print(doc.header, end="", file=output)
    print(doc.title, end="", file=output)
    print(doc.meta, end="", file=output)
    print(doc.inputs, end="", file=output)
    print(doc.outputs, end="", file=output)


def document_task(task: wdl.Task, output: TextIO, keys: Dict[str, str]) -> None:
    doc = MarkDownNode()
    document_node(task, doc, keys)
    print(doc.title, end="", file=output)
    print(doc.meta, end="", file=output)
    print(doc.inputs, end="", file=output)
    print(doc.outputs, end="", file=output)


def parse_file(
    file: str, outdir: str, keys: Dict[str, str], fast_fail: bool = False
) -> None:
    failure = False
    try:
        document = wdl.load(file, read_source=read_source)
    except FileNotFoundError as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True
    except wdl.Error.SyntaxError as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True
    except wdl.Error.ValidationError as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True
    except wdl.Error.RuntimeError as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True
    except wdl.Error.ImportError as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True
    except wdl.Error.MultipleValidationErrors as e:
        logzero.logger.error(f"{e.__class__.__name__}: {e}")
        failure = True

    if fast_fail and failure:
        sys.exit(1)
    elif failure:
        logzero.logger.error(f"Skipping {file}")
        return

    logzero.logger.info(f"Generating docs for {file}")

    outfile_basename = "".join(os.path.basename(file).split(".")[:-1]) + ".md"

    if getattr(document, "workflow"):
        new_filename = os.path.join(outdir, "workflows", outfile_basename)
        pathlib.Path(os.path.dirname(new_filename)).mkdir(parents=True, exist_ok=True)
        output = open(new_filename, "w")
        document_workflow(document, output, keys)
        for task in document.tasks:
            document_task(task, output, keys)
        output.close()
    elif getattr(document, "tasks"):
        new_filename = os.path.join(outdir, "tasks", outfile_basename)
        pathlib.Path(os.path.dirname(new_filename)).mkdir(parents=True, exist_ok=True)
        output = open(new_filename, "w")
        doc = MarkDownNode()
        doc.generate_header(document.source_text)
        print(doc.header, end="", file=output)
        for task in document.tasks:
            document_task(task, output, keys)
        output.close()
    else:
        logzero.logger.warning(f"{file} has no task or workflow definitions.")


def traverse_directory(indir: str, outdir: str, keys: Dict[str, str]) -> None:
    files = glob.glob(indir + f"{os.path.sep}**{os.path.sep}*.wdl", recursive=True)
    for file in files:
        parse_file(file, outdir, keys)



import os
from typing import Any, Optional

import requests
import WDL as wdl
from cachecontrol import CacheControl
from cachecontrol.caches import FileCache
from logzero import logger

# though `forever` is set, this is still invalidated
# if the ETag does not match (meaning the file underneath)
# changed.
sess = CacheControl(
    requests.Session(),
    cache=FileCache(
        os.path.join(os.path.expanduser("~"), ".wdldoc-cache"), forever=True
    ),
)


async def read_source(
    uri: str, path: str, importer: Optional[wdl.Document]
) -> wdl.ReadSourceResult:
    logger.debug(f"Reading source: {uri}")

    if uri.startswith("http:") or uri.startswith("https:"):
        return wdl.ReadSourceResult(sess.get(uri).text, uri)

    return await wdl.read_source_default(uri, path, importer)



from collections import defaultdict
from typing import DefaultDict, Union

import WDL as wdl

from logzero import logger


def classify_inputs(
    node: Union[wdl.Workflow, wdl.Task]
) -> DefaultDict[str, DefaultDict[str, wdl.Env.Binding]]:
    results: DefaultDict[str, DefaultDict[str, wdl.Env.Binding]] = defaultdict(
        lambda: defaultdict(dict)
    )

    for b in reversed(list(node.available_inputs)):
        assert isinstance(b, wdl.Env.Binding)
        var_name, var_value = str(b.name), b.value
        if not b.value.expr and not b.value.type.optional:
            results["required"][var_name] = var_value
        elif b.value.expr and not b.value.type.optional:
            results["default"][var_name] = var_value
        elif b.value.type.optional:
            results["optional"][var_name] = var_value

    return results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wdldoc', 'wdldoc.bin', 'wdldoc.miniwdl', 'wdldoc.templates']

package_data = \
{'': ['*']}

install_requires = \
['cachecontrol[filecache]>=0.12.6,<0.13.0',
 'logzero>=1.5.0,<2.0.0',
 'miniwdl>=0.7.0,<0.8.0',
 'python-semantic-release>=5.1.0,<6.0.0',
 'requests>=2.23.0,<3.0.0']

entry_points = \
{'console_scripts': ['wdldoc = wdldoc.__main__:main']}

setup_kwargs = {
    'name': 'wdldoc',
    'version': '1.7.0',
    'description': 'Create WDL documentation using Markdown.',
    'long_description': '<p align="center">\n  <h1 align="center">\n  wdldoc\n  </h1>\n\n  <p align="center">\n    Convert WDL documentation to Markdown for rendering.\n    <br />\n    <a href="https://github.com/stjudecloud/wdldoc/issues">Request Feature</a>\n    ·\n    <a href="https://github.com/stjudecloud/wdldoc/issues">Report Bug</a>\n    ·\n    ⭐ Consider starring the repo! ⭐\n    <br />\n  </p>\n</p>\n\n## 📚 Getting Started\n\nFor an example of what the results can look like, check out the [GitHub Pages](https://stjudecloud.github.io/workflows/) for the [St Jude Cloud Workflows](https://github.com/stjudecloud/workflows) repo! The documentation is automatically built for each release using `wdldoc`.\n\n### Installation\n\nwdldoc is only available for Python 3.8 or higher.\n\nSuggested install method:\n\n```bash\nconda create -n wdldoc python=3.8\nconda activate wdldoc\npip install wdldoc\n```\n\n## Usage\n\nwdldoc is designed to be simple, and require as little work as possible. Once installed, simply call `wdldoc .` at the root of your WDL project, and Markdown files will be generated in the `./documentation` directory for each WDL file found. There are `tasks/` and `workflows/` subdirectories, with documentation for WDL workflow files in `workflows/`, and documentation for WDL task files in `tasks/`.\n\nAny valid WDL file will have the inputs, outputs, and meta information individually documented for all its tasks and workflows. There\'s no need to conform to any standards we dictate; if it runs, we\'ll document it.\n\nAny strings found in meta fields will be treated as Markdown, so feel free to add custom bolding, italicizing, code snippets, etc.\n\nIf there\'s any information you want to include for a file that doesn\'t fit into a meta field of one of it\'s tasks or workflows, you can include a header section of your WDL file, and we\'ll convert it to Markdown and prepend it to the documentation. This is a good place to document the uses for the file and any licensing information. Simply start a line with `## ` and the rest of the line will be parsed as Markdown. You can include as many header lines you want, and they will be treated as one block. It\'s good practice to break up the header into sections using Markdown titles.\n\n```text\nusage: wdldoc [-h] [-o OUTPUT_DIRECTORY] [-d DESCRIPTION] [-c CHOICES] [-v] [--debug] sources [sources ...]\n\nGenerate clean WDL documentation from source.\n\npositional arguments:\n  sources               Top level directories to search for WDL files, or the WDL files themselves.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -o OUTPUT_DIRECTORY, --output_directory OUTPUT_DIRECTORY\n                        Directory to store markdown files. Default is `./documentation`\n  -d DESCRIPTION, --description DESCRIPTION\n                        If parameter meta fields use a JSON object, the key for the field containing the input description. Default is \'help\'. Ignored if only strings are used.\n  -c CHOICES, --choices CHOICES\n                        If parameter meta fields use a JSON object, the key for the field containing the input choices. Default is \'choices\'. Ignored if only strings are used.\n  -v, --verbose         Sets the log level to INFO.\n  --debug               Sets the log level to DEBUG.\n```\n\nEither directories or individual files can be supplied. When directories are supplied,\nwdldoc will recursively search the input directories searching for all `.wdl` files, and generate documentation for them.\n\nWDL `parameter_meta` info can be anything that conforms to the WDL spec, but we recommend one of two formats. The first is simply `input_name: "descriptive string"`. The other is a JSON object containing a description key with a string value and optionally a choices key with a list of options. The value of the "description" and "choices" keys can be specified with the `--description` and `--choices` arguments. Below is an example of both formats in one parameter meta block.\n\n```text\nparameter_meta {\n    in_bams: {\n        help: "Provide bams to run for comparison"\n    }\n    tissue_type: {\n        help: "Provide the tissue type to compare against",\n        choices: [\'blood\', \'brain\', \'solid\']\n    }\n    output_filename: "Name for the output HTML t-SNE plot"\n}\n```\n\n## 🖥️ Development\n\nIf you are interested in contributing to the code, please first review\nour [CONTRIBUTING.md][contributing-md] document. To bootstrap a\ndevelopment environment, please use the following commands.\n\n```bash\n# Clone the repository\ngit clone git@github.com:stjudecloud/wdldoc.git\ncd wdldoc\n\n# Install the project using poetry\npoetry install\n\n# Ensure pre-commit is installed to automatically format\n# code using `black`.\nbrew install pre-commit\npre-commit install\npre-commit install --hook-type commit-msg\n```\n\n## 📝 License\n\nCopyright © 2020 [St. Jude Cloud Team](https://github.com/stjudecloud).\n\nThis project is [MIT][license-md] licensed.\n\n[contributing-md]: https://github.com/stjudecloud/wdldoc/blob/master/CONTRIBUTING.md\n[license-md]: https://github.com/stjudecloud/wdldoc/blob/master/LICENSE.md\n',
    'author': 'Clay McLeod',
    'author_email': 'Clay.McLeod@STJUDE.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stjudecloud/wdldoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from typing import Any, DefaultDict, Dict, Union

import WDL as wdl


class MarkDownNode:
    def __init__(self) -> None:
        self.title = ""
        self.header = ""
        self.meta = ""
        self.inputs = "\n### Inputs"
        self.outputs = "\n### Outputs"

    def set_title(self, title: str) -> None:
        self.title = "\n## " + title + "\n"

    def generate_header(self, source_text: str) -> None:
        comments = re.findall("^##.*", source_text, re.MULTILINE)
        for comment in comments:
            self.header += comment[3:] + "\n"

    def generate_meta(self, meta: Dict[str, Any]) -> None:
        for key, value in meta.items():
            self.meta += f"\n{key}\n: {value}\n"

    def _parse_description(
        self, description: Union[str, Any], keys: Dict[str, str]
    ) -> None:
        if isinstance(description, str):
            self.inputs += ": {}".format(description)
            return
        if keys["description"] in description:
            self.inputs += ": {}".format(description[keys["description"]])
            del description[keys["description"]]
            if keys["choices"] in description:
                self.inputs += "; **Choices**: {}".format(description[keys["choices"]])
                del description[keys["choices"]]
        for key in description:
            self.inputs += "; **{}**: {}".format(key, description[key])

    def generate_inputs(
        self,
        inputs: DefaultDict[str, DefaultDict[str, wdl.Env.Binding]],
        parameter_metadata: Dict[str, Any],
        keys: Dict[str, str],
    ) -> None:
        if inputs["required"].items():
            self.inputs += "\n\n#### Required\n"
        for name, value in sorted(
            inputs["required"].items(), key=lambda i: (i[0].count("."), i[0])
        ):
            self.inputs += f"\n  * `{name}` ({value.type}, **required**)"
            description = parameter_metadata.get(value.name)
            if description:
                self._parse_description(description, keys)

        if inputs["optional"].items():
            self.inputs += "\n\n#### Optional\n"
        for name, value in sorted(
            inputs["optional"].items(), key=lambda i: (i[0].count("."), i[0])
        ):
            self.inputs += f"\n  * `{name}` ({value.type})"
            description = parameter_metadata.get(value.name)
            if description:
                self._parse_description(description, keys)

        if inputs["default"].items():
            self.inputs += "\n\n#### Defaults\n"
        for name, value in sorted(
            inputs["default"].items(), key=lambda i: (i[0].count("."), i[0])
        ):
            self.inputs += f"\n  * `{name}` ({value.type}, default={value.expr})"
            description = parameter_metadata.get(value.name)
            if description:
                self._parse_description(description, keys)

        if self.inputs == "\n### Inputs":
            self.inputs += "\n**None**\n"
        else:
            self.inputs += "\n"

    def generate_outputs(self, outputs: wdl.Env.Bindings) -> None:
        self.outputs += "\n"
        for output in outputs:
            self.outputs += f"\n  * `{output.name}` ({output.value})"

        if self.outputs == "\n### Outputs\n":
            self.outputs += "**None**\n"
        else:
            self.outputs += "\n"
